require 'spec_helper'

module Contender

  describe FutureTask do
    it 'defers the result of executing a task' do
      result = Object.new
      block = proc do
        result
      end

      task = FutureTask.new block
      task.call

      task.result.should be(result)
    end

    it 'defers the failure of executing a task' do
      block = proc do
        raise ArgumentError
      end

      task = FutureTask.new block
      task.call

      expect {
        task.result
      }.to raise_error(ExecutionError)
    end

    context 'when waiting on the result of the task' do
      it 'raises an exception if the task is not done before the timeout is reached' do
        block = Object.new
        task = FutureTask.new block

        expect {
          task.result 0
        }.to raise_error(TimeoutError)
      end

      it 'raises an exception if the task was cancelled while waiting' do
        start_latch = CountdownLatch.new 1
        cancel_latch = CountdownLatch.new 1

        block = Object.new
        task = FutureTask.new block

        Thread.new do
          start_latch.countdown
          begin
            task.result
          rescue CancellationError
            cancel_latch.countdown
          end
        end

        start_latch.await

        task.cancel false

        cancel_latch.await
      end

      it 'returns the result of the task after it has ran' do
        start_latch = CountdownLatch.new 1

        result = Object.new
        block = proc do
          result
        end

        task = FutureTask.new block

        Thread.new do
          start_latch.countdown
          task.call
        end

        start_latch.await
        task.result.should be(result)
      end
    end

    context 'when cancelling the task' do
      it 'prevents a task from running if it has not already' do
        block = Object.new

        task = FutureTask.new block
        task.cancel false

        expect {
          task.result
        }.to raise_error(CancellationError)

        task.cancelled?.should be_true
      end

      it 'will not cancel a task if it has already finished running' do
        result = Object.new
        block = proc do
          result
        end

        task = FutureTask.new block
        task.call

        task.cancel(false).should be_false
        task.result.should be(result)
      end

      it 'will not interrupt a task if it is running but should_interrupt is false' do
        result = Object.new

        start_latch = CountdownLatch.new 1
        work_latch = CountdownLatch.new 1

        block = proc do
          start_latch.countdown
          work_latch.await
          result
        end

        task = FutureTask.new block

        thread = Thread.new do
          task.call
        end

        start_latch.await

        task.cancel(false).should be_false

        work_latch.countdown

        task.result.should be(result)
      end

      it 'will interrupt a task if it is running and should_interrupt is true' do
        start_latch = CountdownLatch.new 1
        interrupt_latch = CountdownLatch.new 1

        block = proc do
          start_latch.countdown
          begin
            sleep
          rescue InterruptError
            interrupt_latch.countdown
          end
        end

        task = FutureTask.new block

        Thread.new do
          task.call
        end

        start_latch.await

        task.cancel(true).should be_true

        interrupt_latch.await

        expect {
          task.result
        }.to raise_error(CancellationError)
      end
    end
  end

end
